from dataclasses import dataclass
from typing import Optional

from pythonbible.books import Book


@dataclass
class NormalizedReference:
    """NormalizedReference is a dataclass that represents a single scripture reference
    that contains one or more consecutive verses

    :param book: the first book of the Bible in the reference
    :type book: :ref:`Book`
    :param start_chapter: the number of the first chapter in the reference
    :type start_chapter: int
    :param start_verse: the number of the first verse in the reference
    :type start_verse: int
    :param end_chapter: the number of the last chapter in the reference
    :type end_chapter: int
    :param end_verse: the number of the last verse in the reference
    :type end_verse: int
    :param end_book: the last book of the Bible in the reference if the reference
    contains more than one book, defaults to None
    :type end_book: :ref:`Book`
    """

    book: Book
    start_chapter: int
    start_verse: int
    end_chapter: int
    end_verse: int
    end_book: Optional[Book] = None



try:
    import regex as re
except ModuleNotFoundError:
    import re


from typing import Dict, List, Match, Optional, Pattern, Tuple

from .books import Book
from .normalized_reference import NormalizedReference
from .regular_expressions import (
    BOOK_REGULAR_EXPRESSIONS,
    SCRIPTURE_REFERENCE_REGULAR_EXPRESSION,
)
from .roman_numeral_util import convert_all_roman_numerals_to_integers
from .validator import is_valid_reference
from .verses import (
    get_number_of_verses,
    get_number_of_chapters,
    is_single_chapter_book,
)


def get_references(
    text: str, book_groups: Dict[str, List[Book]] = None
) -> List[NormalizedReference]:
    """
    Searches the text for scripture references and returns any that are found in a list of normalized tuple references.

    :param text: String that may contain zero or more scripture references
    :type text: str
    :param book_groups: Optional dictionary of BookGroup (e.g. Old Testament) to its related regular expression
    :type book_groups: Dict[str, List[Book]] or None
    :return: The list of found scripture references
    :rtype: List[NormalizedReference]
    """
    references: List[NormalizedReference] = []

    # First replace all roman numerals in the text with integers.
    clean_text: str = convert_all_roman_numerals_to_integers(text)

    for reference_match in re.finditer(
        SCRIPTURE_REFERENCE_REGULAR_EXPRESSION, clean_text
    ):
        references.extend(normalize_reference(reference_match[0]))

    if book_groups:
        references.extend(_get_book_group_references(clean_text, book_groups))

    return references


def normalize_reference(reference: str) -> List[NormalizedReference]:
    """
    Converts a scripture reference string into a list of normalized tuple references.

    :param reference: a string that is a scripture reference
    :return: a list of tuples. each tuple is in the format (book, start_chapter, start_verse, end_chapter, end_verse)
    """
    references: List[NormalizedReference] = []
    books: List[Book] = []
    cleaned_references: List[str] = []
    reference_without_books: str = reference
    start: int
    end: int
    book_found: bool = True

    while book_found:
        book_found = False

        for book, regular_expression in BOOK_REGULAR_EXPRESSIONS.items():
            reference_match: Optional[Match[str]] = re.search(
                regular_expression, reference_without_books, re.IGNORECASE
            )

            if reference_match:
                start, end = reference_match.regs[0]

                if start != 0 and not books:
                    continue

                book_found = True

                if books:
                    cleaned_references.append(reference_without_books[:start])

                reference_without_books = reference_without_books[end:]
                books.append(book)
                continue

    cleaned_references.append(reference_without_books)

    # First Book
    first_book_references = _process_sub_references(
        books[0], cleaned_references[0].strip()
    )

    if len(books) == 1:
        return first_book_references

    # Second Book
    second_book_references = _process_sub_references(
        books[1], cleaned_references[1].strip()
    )

    if len(first_book_references) > 1:
        references.extend(first_book_references[:-1])

    # Combine last first reference with first second reference
    last_first_reference = first_book_references[-1]
    first_second_reference = second_book_references[0]

    references.append(
        NormalizedReference(
            last_first_reference.book,
            last_first_reference.start_chapter,
            last_first_reference.start_verse,
            first_second_reference.end_chapter,
            first_second_reference.end_verse,
            first_second_reference.book,
        )
    )

    if len(second_book_references) > 1:
        references.extend(second_book_references[1:])

    return references


def _process_sub_references(book: Book, reference: str) -> List[NormalizedReference]:
    references: List[NormalizedReference] = []
    start_chapter: int = 0

    for sub_reference in reference.split(","):
        if (len(sub_reference) == 0 or sub_reference in ["-", "."]) and not references:
            max_chapter: int = get_number_of_chapters(book)
            max_verse: int = get_number_of_verses(book, max_chapter)
            references.append(NormalizedReference(book, 1, 1, max_chapter, max_verse))
            continue

        if sub_reference.endswith("-"):
            sub_reference = sub_reference[:-1]

        start_chapter, start_verse, end_chapter, end_verse = _process_sub_reference(
            sub_reference, book, start_chapter
        )

        new_reference = NormalizedReference(
            book, start_chapter, start_verse, end_chapter, end_verse
        )

        if is_valid_reference(new_reference):
            references.append(new_reference)

        start_chapter = end_chapter

    return references


def _process_sub_reference(
    sub_reference: str, book: Book, start_chapter: int
) -> Tuple[int, int, int, int]:
    start_verse: int = 0
    end_chapter: int = start_chapter
    end_verse: int = start_verse
    no_verses: bool = False

    clean_sub_reference: str = sub_reference.replace(".", ":")
    chapter_and_verse_range: List[str] = clean_sub_reference.split("-")
    min_chapter_and_verse: List[str] = chapter_and_verse_range[0].strip().split(":")

    if len(min_chapter_and_verse) == 1:
        if start_chapter > 0:
            start_verse = int(min_chapter_and_verse[0].strip())
            end_chapter = start_chapter
            end_verse = start_verse
        elif is_single_chapter_book(book):
            start_chapter = 1
            start_verse = int(min_chapter_and_verse[0].strip())
            end_chapter = 1
            end_verse = start_verse
        else:
            start_chapter = int(min_chapter_and_verse[0].strip())
            start_verse = 1
            end_chapter = start_chapter
            end_verse = get_number_of_verses(book, end_chapter)
            no_verses = True
    elif len(min_chapter_and_verse) == 2:
        start_chapter = int(min_chapter_and_verse[0].strip())
        end_chapter = start_chapter
        start_verse = int(min_chapter_and_verse[1].strip())
        end_verse = start_verse

    if len(chapter_and_verse_range) > 1:
        max_chapter_and_verse = chapter_and_verse_range[1].split(":")

        if len(max_chapter_and_verse) == 1:
            if no_verses:
                end_chapter = int(max_chapter_and_verse[0].strip())
                end_verse = get_number_of_verses(book, end_chapter)
            else:
                end_verse = int(max_chapter_and_verse[0].strip())
        elif len(max_chapter_and_verse) == 2:
            end_chapter = int(max_chapter_and_verse[0].strip())
            end_verse = int(max_chapter_and_verse[1].strip())

    return start_chapter, start_verse, end_chapter, end_verse


def _get_book_group_references(
    text: str, book_groups: Dict[str, List[Book]]
) -> List[NormalizedReference]:
    references: List[NormalizedReference] = []
    book_group_regex: Pattern[str] = re.compile(
        "|".join(book_groups.keys()), re.IGNORECASE | re.UNICODE
    )

    for match in re.finditer(book_group_regex, text):
        references.extend(_process_book_group_match(match[0], book_groups))

    return references


def _process_book_group_match(
    text: str, book_groups: Dict[str, List[Book]]
) -> List[NormalizedReference]:
    references: List[NormalizedReference] = []
    books: List[Book] = []

    for regular_expression, books in book_groups.items():
        reference_match: Optional[Match[str]] = re.match(
            regular_expression, text, re.IGNORECASE
        )

        if reference_match:
            break

    start_book: Book = books[0]
    previous_book: Book = start_book

    for book in books[1:]:
        if book.value == previous_book.value + 1:
            previous_book = book
            continue

        references.append(_build_book_group_reference(start_book, previous_book))
        start_book = book
        previous_book = book

    references.append(_build_book_group_reference(start_book, previous_book))

    return references


def _build_book_group_reference(
    start_book: Book, end_book: Book
) -> NormalizedReference:
    max_chapter: int = get_number_of_chapters(end_book)
    max_verse: int = get_number_of_verses(end_book, max_chapter)
    return NormalizedReference(start_book, 1, 1, max_chapter, max_verse, end_book)



from functools import singledispatch
from typing import List

from pythonbible.normalized_reference import NormalizedReference
from pythonbible.parser import get_references


@singledispatch
def count_books(references: List[NormalizedReference]) -> int:
    """
    Returns the count of books of the Bible included in the given list of references.

    :param references: A list of normalized references
    :type references: List[NormalizedReference]
    :return: The count of books of the Bible included in the given list of references
    :rtype: int
    """
    return _get_number_of_books_in_references(references)


@count_books.register
def _count_books_single(reference: NormalizedReference) -> int:
    return _get_number_of_books_in_reference(reference)


@count_books.register
def _count_books_string(reference: str) -> int:
    return _get_number_of_books_in_references(get_references(reference))


def _get_number_of_books_in_references(references: List[NormalizedReference]) -> int:
    return sum(_get_number_of_books_in_reference(reference) for reference in references)


def _get_number_of_books_in_reference(reference: NormalizedReference) -> int:
    return (
        reference.end_book.value - reference.book.value + 1 if reference.end_book else 1
    )



try:
    import regex as re
except ModuleNotFoundError:
    import re


from typing import Dict, Pattern

from pythonbible.books import Book


def build_book_regular_expression(
    book: str, prefix: str = None, suffix: str = None
) -> str:
    return _add_suffix(_add_prefix(book, prefix), suffix)


def _add_prefix(regex: str, prefix: str = None) -> str:
    return regex if prefix is None else rf"(?:{prefix})(?:\s)?{regex}"


def _add_suffix(regex: str, suffix: str = None) -> str:
    return regex if suffix is None else rf"{regex}(?:\s*{suffix})?"


SAMUEL_REGULAR_EXPRESSION = r"(Samuel|Sam\.*|Sa\.*|Sm\.*)"
KINGS_REGULAR_EXPRESSION = r"(Kings|Kgs\.*|Kin\.*|Ki\.*)"
CHRONICLES_REGULAR_EXPRESSION = r"(Chronicles|Chron\.*|Chro\.*|Chr\.*|Ch\.*)"
JOHN_REGULAR_EXPRESSION = r"(John|Joh\.*|Jhn\.*|Jo\.*|Jn\.*)"
CORINTHIANS_REGULAR_EXPRESSION = r"Co\.*(?:r\.*(?:inthians)?)?"
THESSALONIANS_REGULAR_EXPRESSION = r"Th\.*(?:(s|(es(?:s)?))\.*(?:alonians)?)?"
TIMOTHY_REGULAR_EXPRESSION = r"Ti\.*(?:m\.*(?:othy)?)?"
PETER_REGULAR_EXPRESSION = r"(Pe\.*(?:t\.*(?:er)?)?|Pt\.*)"

FIRST = r"1|I\s+|1st\s+|First\s+"
SECOND = r"2|II|2nd\s+|Second\s+"
THIRD = r"3|III|3rd\s+|Third\s+"

FIRST_BOOK = rf"{FIRST}|(First\s+Book\s+of(?:\s+the)?)"
SECOND_BOOK = rf"{SECOND}|(Second\s+Book\s+of(?:\s+the)?)"

EPISTLE_OF_PAUL_TO = r"Epistle\s+of\s+Paul\s+(?:the\s+Apostle\s+)?to(?:\s+the)?"
GENERAL_EPISTLE_OF = r"(?:General\s+)?Epistle\s+(?:General\s+)?of"

FIRST_PAUL_EPISTLE = rf"{FIRST}|(First\s+{EPISTLE_OF_PAUL_TO})"
SECOND_PAUL_EPISTLE = rf"{SECOND}|(Second\s+{EPISTLE_OF_PAUL_TO})"

FIRST_GENERAL_EPISTLE = rf"{FIRST}|(First\s+{GENERAL_EPISTLE_OF})"
SECOND_GENERAL_EPISTLE = rf"{SECOND}|(Second\s+{GENERAL_EPISTLE_OF})"
THIRD_GENERAL_EPISTLE = rf"{THIRD}|(Third\s+{GENERAL_EPISTLE_OF})"

# noinspection SpellCheckingInspection
BOOK_REGULAR_EXPRESSIONS: Dict[Book, str] = {
    Book.GENESIS: r"Gen\.*(?:esis)?",
    Book.EXODUS: r"Exo\.*(?:d\.*)?(?:us)?",
    Book.LEVITICUS: r"Lev\.*(?:iticus)?",
    Book.NUMBERS: r"Num\.*(?:bers)?",
    Book.DEUTERONOMY: r"Deu\.*(?:t\.*)?(?:eronomy)?",
    Book.JOSHUA: r"(Joshua|Josh\.*|Jos\.*|Jsh\.*)",
    Book.JUDGES: r"(Judges|Judg\.*|Jdgs\.*|Jdg\.*)",
    Book.RUTH: r"(Ruth|Rut\.*|Rth\.*)",
    Book.SAMUEL_1: build_book_regular_expression(
        SAMUEL_REGULAR_EXPRESSION,
        prefix=FIRST_BOOK,
        suffix=r"Otherwise\s+Called\s+The\s+First\s+Book\s+of\s+the\s+Kings",
    ),
    Book.SAMUEL_2: build_book_regular_expression(
        SAMUEL_REGULAR_EXPRESSION,
        prefix=SECOND_BOOK,
        suffix=r"Otherwise\s+Called\s+The\s+Second\s+Book\s+of\s+the\s+Kings",
    ),
    Book.KINGS_1: build_book_regular_expression(
        KINGS_REGULAR_EXPRESSION,
        prefix=FIRST_BOOK,
        suffix=r"\,\s+Commonly\s+Called\s+the\s+Third\s+Book\s+of\s+the\s+Kings",
    ),
    Book.KINGS_2: build_book_regular_expression(
        KINGS_REGULAR_EXPRESSION,
        prefix=SECOND_BOOK,
        suffix=r"\,\s+Commonly\s+Called\s+the\s+Fourth\s+Book\s+of\s+the\s+Kings",
    ),
    Book.CHRONICLES_1: build_book_regular_expression(
        CHRONICLES_REGULAR_EXPRESSION,
        prefix=FIRST_BOOK,
    ),
    Book.CHRONICLES_2: build_book_regular_expression(
        CHRONICLES_REGULAR_EXPRESSION,
        prefix=SECOND_BOOK,
    ),
    Book.EZRA: r"Ezr\.*(?:a)?",
    Book.NEHEMIAH: r"Neh\.*(?:emiah)?",
    Book.ESTHER: r"Est\.*(?:h\.*)?(?:er)?",
    Book.JOB: "Job",
    Book.PSALMS: r"(Psalms|Psalm|Pslm\.*|Psa\.*|Psm\.*|Pss\.*|Ps\.*)",
    Book.PROVERBS: r"(Proverbs|Prov\.*|Pro\.*|Prv\.*)",
    Book.ECCLESIASTES: r"(Ecclesiastes(?:\s+or\,\s+the\s+Preacher)?|Eccles\.*|Eccle\.*|Eccl\.*|Ecc\.*|Ec\.*|Qoh\.*)",
    Book.SONG_OF_SONGS: r"(Song(?: of (Solomon|Songs|Sol\.*))?)|Canticles|(Canticle(?: of Canticles)?)|SOS|Cant",
    Book.ISAIAH: r"Isa\.*(?:iah)?",
    Book.JEREMIAH: r"Jer\.*(?:emiah)?",
    Book.LAMENTATIONS: build_book_regular_expression(
        r"Lam\.*(?:entations)?", suffix=r"of\s+Jeremiah"
    ),
    Book.EZEKIEL: r"(Ezekiel|Ezek\.*|Eze\.*|Ezk\.*)",
    Book.DANIEL: r"Dan\.*(?:iel)?",
    Book.HOSEA: r"Hos\.*(?:ea)?",
    Book.JOEL: r"Joe\.*(?:l)?",
    Book.AMOS: r"Amo\.*(?:s)?",
    Book.OBADIAH: r"Oba\.*(?:d\.*(?:iah)?)?",
    Book.JONAH: r"Jonah|Jon\.*|Jnh\.*",
    Book.MICAH: r"Mic\.*(?:ah)?",
    Book.NAHUM: r"Nah\.*(?:um)?",
    Book.HABAKKUK: r"Hab\.*(?:akkuk)?",
    Book.ZEPHANIAH: r"Zep\.*(?:h\.*(?:aniah)?)?",
    Book.HAGGAI: r"Hag\.*(?:gai)?",
    Book.ZECHARIAH: r"Zec\.*(?:h\.*(?:ariah)?)?",
    Book.MALACHI: r"Mal\.*(?:achi)?",
    Book.MATTHEW: r"Mat\.*(?:t\.*(?:hew)?)?",
    Book.MARK: r"Mark|Mar\.*|Mrk\.*",
    Book.LUKE: r"Luk\.*(?:e)?",
    Book.JOHN: r"(?<!(?:1|2|3|I)\s)(?<!(?:1|2|3|I))" + JOHN_REGULAR_EXPRESSION,
    Book.ACTS: build_book_regular_expression(r"Act\.*(?:s)?", suffix="of the Apostles"),
    Book.ROMANS: r"Rom\.*(?:ans)?",
    Book.CORINTHIANS_1: build_book_regular_expression(
        CORINTHIANS_REGULAR_EXPRESSION, prefix=FIRST_PAUL_EPISTLE
    ),
    Book.CORINTHIANS_2: build_book_regular_expression(
        CORINTHIANS_REGULAR_EXPRESSION, prefix=SECOND_PAUL_EPISTLE
    ),
    Book.GALATIANS: r"Gal\.*(?:atians)?",
    Book.EPHESIANS: r"Eph\.*(?:es\.*(?:ians)?)?",
    Book.PHILIPPIANS: r"Ph(?:(p\.*)|(?:il\.*(?!e\.*(?:m\.*(?:on)?)?)(?:ippians)?))",
    Book.COLOSSIANS: r"Col\.*(?:ossians)?",
    Book.THESSALONIANS_1: build_book_regular_expression(
        THESSALONIANS_REGULAR_EXPRESSION, prefix=FIRST_PAUL_EPISTLE
    ),
    Book.THESSALONIANS_2: build_book_regular_expression(
        THESSALONIANS_REGULAR_EXPRESSION, prefix=SECOND_PAUL_EPISTLE
    ),
    Book.TIMOTHY_1: build_book_regular_expression(
        TIMOTHY_REGULAR_EXPRESSION, prefix=FIRST_PAUL_EPISTLE
    ),
    Book.TIMOTHY_2: build_book_regular_expression(
        TIMOTHY_REGULAR_EXPRESSION, prefix=SECOND_PAUL_EPISTLE
    ),
    Book.TITUS: r"Tit\.*(?:us)?",
    # assume 'Phi' is Philemon if Philippians failed
    Book.PHILEMON: r"(Philemon|Philem\.*|Phile\.*|Phlm\.*|Phi\.*|Phm\.*)",
    Book.HEBREWS: r"Heb\.*(?:rews)?",
    Book.JAMES: r"Ja(?:me)?s\.*",
    Book.PETER_1: build_book_regular_expression(
        PETER_REGULAR_EXPRESSION, prefix=FIRST_GENERAL_EPISTLE
    ),
    Book.PETER_2: build_book_regular_expression(
        PETER_REGULAR_EXPRESSION, prefix=SECOND_GENERAL_EPISTLE
    ),
    Book.JOHN_1: build_book_regular_expression(
        JOHN_REGULAR_EXPRESSION, prefix=FIRST_GENERAL_EPISTLE
    ),
    Book.JOHN_2: build_book_regular_expression(
        JOHN_REGULAR_EXPRESSION, prefix=SECOND_GENERAL_EPISTLE
    ),
    Book.JOHN_3: build_book_regular_expression(
        JOHN_REGULAR_EXPRESSION, prefix=THIRD_GENERAL_EPISTLE
    ),
    Book.JUDE: r"Jud\.*(:?e)?",
    Book.REVELATION: build_book_regular_expression(
        r"Rev\.*(?:elation)?", suffix="of ((Jesus Christ)|John|(St. John the Divine))"
    ),
    # Book.ESDRAS_1: None,
    # Book.TOBIT: None,
    # Book.WISDOM_OF_SOLOMON: None,
    # Book.ECCLESIASTICUS: None,
    # Book.MACCABEES_1: None,
    # Book.MACCABEES_2: None,
}

BOOK_REGEX: str = "|".join(
    rf"\b{value}\b\.?" for value in BOOK_REGULAR_EXPRESSIONS.values()
)

CHAPTER_REGEX: str = r"(\d{1,3})"
CHAPTER_VERSE_SEPARATOR: str = r"([:.])"
VERSE_REGEX: str = r"(\d{1,3})"
CHAPTER_AND_VERSE_REGEX: str = (
    rf"({CHAPTER_REGEX}(\s*{CHAPTER_VERSE_SEPARATOR}\s*{VERSE_REGEX})?)"
)
RANGE_REGEX: str = rf"({CHAPTER_AND_VERSE_REGEX}(\s*-\s*({CHAPTER_REGEX}\s*{CHAPTER_VERSE_SEPARATOR}\s*)?{VERSE_REGEX})?)"
ADDITIONAL_REFERENCE_REGEX: str = rf"(\s*,\s*({RANGE_REGEX}|{VERSE_REGEX}))"
FULL_CHAPTER_AND_VERSE_REGEX: str = f"({RANGE_REGEX}({ADDITIONAL_REFERENCE_REGEX})*)"

FULL_BOOK_REGEX = rf"({BOOK_REGEX})\s*({FULL_CHAPTER_AND_VERSE_REGEX})?"
CROSS_BOOK_REGEX = rf"({FULL_BOOK_REGEX}(\s*-\s*({FULL_BOOK_REGEX}))?)"

SCRIPTURE_REFERENCE_REGULAR_EXPRESSION: Pattern[str] = re.compile(
    CROSS_BOOK_REGEX, re.IGNORECASE | re.UNICODE
)



from typing import List

import pytest

import pythonbible as bible
from pythonbible.verses import VERSE_IDS


def test_convert_reference_to_verse_ids(reference: bible.NormalizedReference) -> None:
    # Given a valid normalized scripture reference
    # When the reference is converted into a list of verse id integers
    verse_ids: List[int] = bible.convert_reference_to_verse_ids(reference)

    # Then the resulting list of verse id integers is accurate
    assert len(verse_ids) == 60
    assert verse_ids[0] == 1001001
    assert verse_ids[59] == 1003004


def test_convert_reference_to_verse_ids_null() -> None:
    # Given a null reference
    # When we attempt to convert it into a list of verse ids
    verse_ids: List[int] = bible.convert_reference_to_verse_ids(None)

    # Then the result is an empty list
    assert len(verse_ids) == 0


def test_convert_reference_to_verse_ids_invalid(
    invalid_reference: bible.NormalizedReference,
) -> None:
    # Given an invalid reference
    # When we attempt to convert it into a list of verse ids
    with pytest.raises(bible.InvalidChapterError):
        bible.convert_reference_to_verse_ids(invalid_reference)


def test_convert_references_to_verse_ids(
    references: List[bible.NormalizedReference], verse_ids: List[int]
) -> None:
    # Given a list of valid normalized scripture references
    # When the references are converted into a list of verse id integers
    actual_verse_ids: List[int] = bible.convert_references_to_verse_ids(references)

    # Then the resulting list of verse id integers is accurate
    assert actual_verse_ids == verse_ids


def test_convert_references_to_verse_ids_null() -> None:
    # Given a null references object
    # When we attempt to convert it into a list of verse ids
    actual_verse_ids: List[int] = bible.convert_references_to_verse_ids(None)

    # Then the result is an empty list
    assert len(actual_verse_ids) == 0


def test_convert_references_to_verse_ids_complex(
    normalized_references_complex: List[bible.NormalizedReference],
    verse_ids_complex: List[int],
) -> None:
    # Given a list of complex references
    # When converted into verse ids
    actual_verse_ids: List[int] = bible.convert_references_to_verse_ids(
        normalized_references_complex
    )

    # Then the list of verse ids is correct
    assert actual_verse_ids == verse_ids_complex


def test_convert_verse_ids_to_references(
    verse_ids: List[int], references: List[bible.NormalizedReference]
) -> None:
    # Given a list of integer verse ids
    # When we convert them into a list of normalized reference tuples
    actual_references: List[
        bible.NormalizedReference
    ] = bible.convert_verse_ids_to_references(verse_ids)

    # Then the resulting list of references is accurate
    assert actual_references == references


def test_convert_verse_ids_to_references_null() -> None:
    # Given a null verse_ids object
    # When we attempt to convert them into a list of references
    actual_references: List[
        bible.NormalizedReference
    ] = bible.convert_verse_ids_to_references(None)

    # Then the list of references is empty
    assert len(actual_references) == 0


def test_convert_verse_ids_to_references_invalid(invalid_verse_id: int) -> None:
    # Given a list of verse ids with an invalid verse id
    # When we attempt to convert them into a list of references
    # Then an error is raised
    with pytest.raises(bible.InvalidVerseError, match="1100100 is not a valid verse."):
        bible.convert_verse_ids_to_references([invalid_verse_id])


def test_convert_verse_ids_to_references_invalid2(
    verse_id: int, invalid_verse_id: int
) -> None:
    # Given a list of verse ids with an invalid verse id
    # When we attempt to convert them into a list of references
    # Then an error is raised
    with pytest.raises(bible.InvalidVerseError, match="1100100 is not a valid verse."):
        bible.convert_verse_ids_to_references([verse_id, invalid_verse_id])


def test_convert_verse_ids_to_references_complex(
    normalized_references_complex: List[bible.NormalizedReference],
    verse_ids_complex: List[int],
) -> None:
    # Given a list of "complex" verse ids
    # When we convert them into a list of references
    actual_references: List[
        bible.NormalizedReference
    ] = bible.convert_verse_ids_to_references(verse_ids_complex)

    # Then the list of references is correct
    assert actual_references == normalized_references_complex


def test_whole_book() -> None:
    """Test for https://github.com/avendesora/pythonbible/issues/7!"""
    # Given a reference that is just a book title
    reference_string: str = "Genesis"

    # When we convert that to normalized references
    references: List[bible.NormalizedReference] = bible.get_references(reference_string)

    # Then it should return the normalized reference for the entire book.
    assert len(references) == 1
    assert references[0] == bible.NormalizedReference(bible.Book.GENESIS, 1, 1, 50, 26)


def test_cross_book() -> None:
    # Given a reference that spans multiple books
    reference_string: str = "Paul's epistles"

    # When we convert that to normalized references
    references: List[bible.NormalizedReference] = bible.get_references(
        reference_string, book_groups=bible.BOOK_GROUPS
    )

    # and then convert that to verse ids
    verse_ids: List[int] = bible.convert_references_to_verse_ids(references)

    # Then it return the verse ids for the verses in all of the books in the reference.
    first_verse: int = VERSE_IDS.index(45001001)
    last_verse: int = VERSE_IDS.index(57001025)
    assert verse_ids == VERSE_IDS[first_verse : last_verse + 1]


def test_cross_book_reverse() -> None:
    # Given a list of verse ids that spans multiple books
    first_verse: int = VERSE_IDS.index(45001001)
    last_verse: int = VERSE_IDS.index(57001025)
    verse_ids: List[int] = VERSE_IDS[first_verse : last_verse + 1]

    # When we convert that to normalized references
    references: List[bible.NormalizedReference] = bible.convert_verse_ids_to_references(
        verse_ids
    )

    # Then the result is a single normalized reference that spans multiple books
    # rather than one for each book.
    assert references == [
        bible.NormalizedReference(bible.Book.ROMANS, 1, 1, 1, 25, bible.Book.PHILEMON),
    ]



import time
from typing import List, Optional

import pytest

import pythonbible as bible
from pythonbible.formatter import BookTitles


def test_format_scripture_references(
    normalized_references_complex: List[bible.NormalizedReference],
    formatted_reference: str,
) -> None:
    # Given a list of normalized references
    # When we format them into a reference string
    reference: str = bible.format_scripture_references(normalized_references_complex)

    # Then the reference string is correctly formatted.
    assert reference == formatted_reference


def test_format_scripture_references_null() -> None:
    # Given a null references object
    # When we attempt to format it into a reference string
    formatted_reference: str = bible.format_scripture_references(None)

    # Then the result is null
    assert formatted_reference == ""


def test_format_scripture_references_sorting(
    normalized_references_complex: List[bible.NormalizedReference],
    formatted_reference: str,
) -> None:
    # Given a list of normalized references that are not in proper order
    normalized_references_complex.reverse()

    # When we format them into a reference string
    reference: str = bible.format_scripture_references(normalized_references_complex)

    # Then the reference string is correctly formatted.
    assert reference == formatted_reference


def test_format_scripture_reference_single_verse(verse_id: int) -> None:
    # Given a single verse id
    book: bible.Book
    chapter: int
    verse: int
    book, chapter, verse = bible.get_book_chapter_verse(verse_id)

    # When we format the reference for it for different versions
    short_kjv_reference: str = bible.format_single_reference(
        bible.NormalizedReference(book, chapter, verse, chapter, verse),
        version=bible.Version.KING_JAMES,
    )
    short_asv_reference: str = bible.format_single_reference(
        bible.NormalizedReference(book, chapter, verse, chapter, verse),
        version=bible.Version.KING_JAMES,
    )
    long_kjv_reference: str = bible.format_single_reference(
        bible.NormalizedReference(book, chapter, verse, chapter, verse),
        version=bible.Version.KING_JAMES,
        full_title=True,
    )
    long_asv_reference: str = bible.format_single_reference(
        bible.NormalizedReference(book, chapter, verse, chapter, verse),
        version=bible.Version.AMERICAN_STANDARD,
        full_title=True,
    )

    # Then the short references should match, but the long ones should be different.
    assert short_kjv_reference == short_asv_reference
    assert long_kjv_reference != long_asv_reference


def test_format_scripture_text(verse_ids: List[int], html_scripture_text: str) -> None:
    # Given a list of verse ids
    # When we get the scripture text for those verse ids
    scripture_text: str = bible.format_scripture_text(verse_ids)

    # Then the scripture text is formatted correctly.
    assert scripture_text == html_scripture_text


def test_format_scripture_text_non_html(
    verse_ids: List[int], non_html_scripture_text: str
) -> None:
    # Given a list of verse ids
    # When we get the non html scripture text for those verse ids
    scripture_text: str = bible.format_scripture_text(verse_ids, format_type="text")

    # Then the scripture text is formatted correctly.
    assert scripture_text == non_html_scripture_text


def test_format_scripture_text_one_verse_per_paragraph(
    verse_ids_multiple_chapters: List[int],
    html_scripture_text_one_verse_per_paragraph: str,
) -> None:
    # Given a list of verse ids
    # When we get the scripture text for those verse ids
    scripture_text: str = bible.format_scripture_text(
        verse_ids_multiple_chapters, one_verse_per_paragraph=True
    )

    # Then the scripture text is formatted correctly.
    assert scripture_text == html_scripture_text_one_verse_per_paragraph


def test_get_verse_text(verse_id: int, verse_text_no_verse_number: str) -> None:
    # Given a valid verse id
    # When using that verse to get the verse text
    verse_text: Optional[str] = bible.get_verse_text(
        verse_id, version=bible.Version.KING_JAMES
    )

    # Then the verse text is the appropriate verse text.
    assert verse_text == verse_text_no_verse_number


def test_get_verse_text_invalid(invalid_verse_id: int) -> None:
    # Given an invalid verse id
    # When attempting to get the verse text for that verse id
    # Then an error is raised.
    with pytest.raises(bible.InvalidVerseError, match="1100100 is not a valid verse."):
        bible.get_verse_text(invalid_verse_id)


def test_get_verse_text_no_version_file(verse_id: int) -> None:
    # Given a valid verse id and a version that doesn't have a file
    version: bible.Version = bible.Version.MESSAGE

    # When using that verse id and version to the get the verse text
    # Then a MissingVerseFileError is raised.
    with pytest.raises(bible.MissingVerseFileError):
        bible.get_verse_text(verse_id, version=version)


def test_verse_text_caching() -> None:
    # Given a lengthy reference
    references: List[bible.NormalizedReference] = bible.get_references("Jeremiah 29")
    verse_ids: List[int] = bible.convert_references_to_verse_ids(references)

    # When getting the scripture text multiple times
    first_start_time: float = time.time()
    first_verses: List[Optional[str]] = [
        bible.get_verse_text(verse_id) for verse_id in verse_ids
    ]
    second_start_time: float = time.time()
    second_verses: List[Optional[str]] = [
        bible.get_verse_text(verse_id) for verse_id in verse_ids
    ]
    end_time: float = time.time()

    first_time: float = second_start_time - first_start_time
    second_time: float = end_time - second_start_time

    # Then the results are cached, so we get the same results much faster the second time
    assert first_time * 0.1 > second_time
    assert first_verses == second_verses


def test_get_book_titles(
    book: bible.Book, long_book_title: str, short_book_title: str
) -> None:
    # Given a book
    # When we get the book titles for that book
    book_titles: Optional[BookTitles] = bible.get_book_titles(
        book, version=bible.Version.KING_JAMES
    )

    # Then the long and short book titles match what is expected.
    assert book_titles is not None
    assert book_titles.long_title == long_book_title
    assert book_titles.short_title == short_book_title


def test_get_book_titles_no_version_file(book: bible.Book) -> None:
    # Given a valid book and a version that doesn't have a file
    version: bible.Version = bible.Version.MESSAGE

    # When using that book and version to the get the book titles
    # Then a MissingBookFileError is raised.
    with pytest.raises(bible.MissingBookFileError):
        bible.get_book_titles(book, version)


def test_format_scripture_references_multiple_book_range() -> None:
    # Given a reference that spans multiple books
    references: List[bible.NormalizedReference] = bible.get_references(
        "Old Testament", book_groups=bible.BOOK_GROUPS
    )

    # When formatting that reference into a reference string
    reference_string: str = bible.format_scripture_references(references)

    # Then the resulting reference string should be a range that spans multiple books.
    assert reference_string == "Genesis - Malachi"


def test_single_chapter_books() -> None:
    # Given a reference for a book that has only one chapter
    references: List[bible.NormalizedReference] = bible.get_references("Obadiah 1:2-4")

    # When formatting that reference into a reference string
    reference_string: str = bible.format_scripture_references(references)

    # Then the resulting reference includes the book title and verse numbers but not chapter numbers.
    assert reference_string == "Obadiah 2-4"


def test_single_chapter_book_in_multiple_book_reference() -> None:
    # Given a reference that includes a book with only one chapter but spans multiple books
    references: List[bible.NormalizedReference] = bible.get_references(
        "Amos 1:3 - Obadiah 1:12"
    )

    # When formatting that reference into a reference string
    reference_string: str = bible.format_scripture_references(references)

    # Then the resulting reference has chapter numbers for the book with multiple
    # chapters and no chapter numbers for the single chapter book.
    assert reference_string == "Amos 1:3 - Obadiah 12"


def test_single_chapter_books_force_chapter_numbers() -> None:
    # Given a reference for a book that has only one chapter
    references: List[bible.NormalizedReference] = bible.get_references("Obadiah")

    # When formatting that reference into a reference string and including the keyword
    # argument to force include chapter numbers
    reference_string: str = bible.format_scripture_references(
        references, always_include_chapter_numbers=True
    )

    # Then the resulting reference includes the book title and verse numbers but not chapter numbers.
    assert reference_string == "Obadiah 1:1-21"


def test_cross_book_range_not_whole_books_or_chapters() -> None:
    # Given a reference that is a range that spans multiple books but not entire books/chapters
    original_reference_string = "Genesis 50:3 - Exodus 1:10"
    references: List[bible.NormalizedReference] = bible.get_references(
        original_reference_string
    )

    # When formatting that reference into a reference string
    reference_string: str = bible.format_scripture_references(references)

    # Then the resulting reference string includes chapter and verse numbers
    assert reference_string == original_reference_string


def test_multi_chapter_book_reference_contains_whole_book() -> None:
    # Given a reference that contains all the verses of a single book that has multiple chapters
    references: List[bible.NormalizedReference] = bible.get_references(
        "Genesis 1:1-50:26"
    )

    # When formatting that reference into a reference string
    reference_string: str = bible.format_scripture_references(references)

    # Then the resulting reference string does not include chapter and verse numbers
    assert reference_string == "Genesis"



from typing import List

from pythonbible.books import Book
from pythonbible.errors import InvalidVerseError
from pythonbible.normalized_reference import NormalizedReference
from pythonbible.validator import is_valid_verse_id
from pythonbible.verses import VERSE_IDS, get_book_chapter_verse, get_verse_id


def convert_references_to_verse_ids(references: List[NormalizedReference]) -> List[int]:
    """
    Converts the given list of NormalizedReference objects into a list of verse id integers.

    :param references: A list of normalized references
    :type references: List[NormalizedReference]
    :return: The list of verse ids associated with the references
    :rtype: List[int]
    """
    verse_ids: List[int] = []

    if references is not None:
        for reference in references:
            verse_ids.extend(convert_reference_to_verse_ids(reference))

    return verse_ids


def convert_reference_to_verse_ids(reference: NormalizedReference) -> List[int]:
    """
    Converts the given NormalizedReference object into a list of verse id integers.

    :param reference: A normalized reference
    :type reference: NormalizedReference
    :return: The list of verse ids associated with the reference
    :rtype: List[int]
    """
    if reference is None:
        return []

    end_book = reference.book if reference.end_book is None else reference.end_book

    start_verse_id: int = get_verse_id(
        reference.book, reference.start_chapter, reference.start_verse
    )
    end_verse_id: int = get_verse_id(
        end_book, reference.end_chapter, reference.end_verse
    )
    return VERSE_IDS[
        VERSE_IDS.index(start_verse_id) : VERSE_IDS.index(end_verse_id) + 1
    ]


def convert_verse_ids_to_references(verse_ids: List[int]) -> List[NormalizedReference]:
    """
    Converts the given list of verse id integers into a list of NormalizedReference objects.

    :param verse_ids: A list of verse ids
    :type verse_ids: List[int]
    :return: The list of normalized references associated with the verse ids
    :rtype: List[NormalizedReference]
    :raises InvalidVerseError: if one or more of the verse_ids does not correspond to a valid verse
    """
    references: List[NormalizedReference] = []

    if verse_ids is None or not verse_ids:
        return references

    verse_ids.sort()

    # Initialize with the first verse id in the list
    first_verse = verse_ids[0]

    if not is_valid_verse_id(first_verse):
        raise InvalidVerseError(verse_id=first_verse)

    previous_verse_id: int = verse_ids[0]

    book: Book
    chapter: int
    verse: int
    book, chapter, verse = get_book_chapter_verse(previous_verse_id)

    start_book: Book = book
    previous_book: Book = book
    start_chapter: int = chapter
    start_verse: int = verse
    previous_chapter: int = chapter
    previous_verse: int = verse

    # Loop through the remaining verse ids in the list
    for verse_id in verse_ids[1:]:
        if not is_valid_verse_id(verse_id):
            raise InvalidVerseError(verse_id=verse_id)

        book, chapter, verse = get_book_chapter_verse(verse_id)

        # If it's just the next verse in the range, updated the previous fields and continue.
        if VERSE_IDS.index(verse_id) - VERSE_IDS.index(previous_verse_id) == 1:
            previous_book = book
            previous_chapter = chapter
            previous_verse = verse
            previous_verse_id = verse_id
            continue

        # At the beginning of a new range, so create the reference and reset all of the fields.
        references.append(
            NormalizedReference(
                start_book,
                start_chapter,
                start_verse,
                previous_chapter,
                previous_verse,
                previous_book if start_book != previous_book else None,
            )
        )

        start_book = book
        previous_book = book
        start_chapter = chapter
        start_verse = verse
        previous_chapter = chapter
        previous_verse = verse
        previous_verse_id = verse_id

    # The last range reference doesn't get created within the loop, so create it now.
    references.append(
        NormalizedReference(
            start_book,
            start_chapter,
            start_verse,
            previous_chapter,
            previous_verse,
            previous_book if start_book != previous_book else None,
        )
    )

    return references



import json
import os
from dataclasses import dataclass
from functools import lru_cache
from typing import Any, Callable, Dict, List, Optional, Type

from pythonbible.bible.bible_parser import BibleParser
from pythonbible.bible.osis.parser import OSISParser
from pythonbible.books import Book
from pythonbible.converter import (
    convert_references_to_verse_ids,
    convert_verse_ids_to_references,
)
from pythonbible.errors import (
    InvalidVerseError,
    MissingBookFileError,
    MissingVerseFileError,
)
from pythonbible.normalized_reference import NormalizedReference
from pythonbible.verses import (
    VERSE_IDS,
    get_book_chapter_verse,
    get_number_of_verses,
    get_number_of_chapters,
    is_single_chapter_book,
)
from pythonbible.versions import DEFAULT_VERSION, Version


@dataclass
class BookTitles:
    long_title: str
    short_title: str


VERSION_MAP: Dict[Version, Type[BibleParser]] = {
    Version.AMERICAN_STANDARD: OSISParser,
    Version.KING_JAMES: OSISParser,
}


def get_parser(**kwargs) -> BibleParser:
    version: Version = kwargs.get("version", DEFAULT_VERSION)
    version_map: Dict[Version, Type[BibleParser]] = kwargs.get(
        "version_map", VERSION_MAP
    )
    return version_map.get(version, OSISParser)(version)


DEFAULT_PARSER: BibleParser = get_parser()

CURRENT_FOLDER: str = os.path.dirname(os.path.realpath(__file__))
DATA_FOLDER: str = os.path.join(os.path.join(CURRENT_FOLDER, "bible"), "data")
VERSE_TEXTS: Dict[Version, Dict[int, str]] = {}
BOOK_TITLES: Dict[Version, Dict[Book, BookTitles]] = {}


# TODO - handle Psalms vs Psalm appropriately
# TODO - handle single chapter books appropriately (e.g. Obadiah 1-4 rather than Obadiah 1:1-4)
def format_scripture_references(
    references: Optional[List[NormalizedReference]], **kwargs
) -> str:
    """
    Returns a human-readable string of the given normalized scripture references

    :param references: A list of normalized scripture references
    :type references: List[NormalizedReference]
    :return: A human-readable string of the given normalized scripture references
    :rtype: str
    """
    if references is None:
        return ""

    sorted_references: List[NormalizedReference] = references

    # Only sort if there is more than one reference as it can take a long time if there
    # are a lot of verses covered by the references.
    if len(references) > 1:
        verse_ids: List[int] = convert_references_to_verse_ids(references)
        verse_ids.sort()
        sorted_references = convert_verse_ids_to_references(verse_ids)

    formatted_reference: str = ""

    previous_reference: Optional[NormalizedReference] = None

    for reference in sorted_references:
        previous_book: Optional[Book] = _get_previous_book(previous_reference)

        if previous_book != reference.book:
            if previous_reference:
                formatted_reference += ";"

            formatted_reference += format_single_reference(reference, **kwargs)
            previous_reference = reference
            continue

        if _is_reference_with_a_new_chapter(previous_reference, reference):
            formatted_reference += ","
            formatted_reference += format_single_reference(
                reference, include_books=False, **kwargs
            )
            continue

        # Reference with same book and chapter as previous reference
        formatted_reference += ","
        formatted_reference += format_single_reference(
            reference, include_books=False, include_chapters=False, **kwargs
        )
        previous_reference = reference

    return formatted_reference


def _get_previous_book(reference: Optional[NormalizedReference]) -> Optional[Book]:
    if reference is None:
        return None

    return reference.book if reference.end_book is None else reference.end_book


def _is_reference_with_a_new_chapter(
    previous_reference: Optional[NormalizedReference],
    current_reference: NormalizedReference,
) -> bool:
    if (
        previous_reference
        and previous_reference.end_chapter != current_reference.start_chapter
    ):
        return True

    return current_reference.end_chapter > current_reference.start_chapter


def format_single_reference(
    reference: NormalizedReference,
    include_books: bool = True,
    include_chapters: bool = True,
    **kwargs,
) -> str:
    """
    Returns a human-readable string of the given normalized scripture reference

    :param reference: A normalized scripture reference
    :type reference: NormalizedReference
    :param include_books: If True includes the book title(s) in the returned reference string, defaults to True
    :type include_books: bool
    :param include_chapters: If True includes the chapter number(s) in the returned reference string, defaults to True
    :type include_chapters: bool
    :return: A human-readable string of the given normalized scripture reference
    :rtype: str
    """
    start_book: str = _get_start_book(reference, include_books, **kwargs)
    start_chapter: str = _get_start_chapter(reference, include_chapters, **kwargs)
    start_verse: str = _get_start_verse(reference, **kwargs)
    end_book: str = _get_end_book(reference, include_books, **kwargs)
    end_chapter: str = _get_end_chapter(reference, include_chapters, **kwargs)
    end_verse: str = _get_end_verse(reference, **kwargs)

    start_separator: str = " " if start_book and (start_chapter or start_verse) else ""
    end_separator: str = " " if end_book and (end_chapter or end_verse) else ""
    range_separator: str = (
        " - " if end_book else "-" if end_chapter or end_verse else ""
    )

    return "".join(
        [
            start_book,
            start_separator,
            start_chapter,
            start_verse,
            range_separator,
            end_book,
            end_separator,
            end_chapter,
            end_verse,
        ]
    )


def _get_start_book(
    reference: NormalizedReference, include_books: bool = True, **kwargs
) -> str:
    return _get_book_title(reference.book, include_books, **kwargs)


def _get_end_book(
    reference: NormalizedReference, include_books: bool = True, **kwargs
) -> str:
    if reference.end_book and reference.end_book != reference.book:
        return _get_book_title(reference.end_book, include_books, **kwargs)

    return ""


def _get_book_title(book: Book, include_books: bool = True, **kwargs) -> str:
    if not include_books:
        return ""

    version: Version = kwargs.get("version", DEFAULT_VERSION)
    full_title: bool = kwargs.get("full_title", False)
    book_titles: Optional[BookTitles] = get_book_titles(book, version)

    return (
        (book_titles.long_title if full_title else book_titles.short_title)
        if book_titles
        else ""
    )


def _get_start_chapter(
    reference: NormalizedReference, include_chapters: bool = True, **kwargs
) -> str:
    if not include_chapters:
        return ""

    force_include_chapters: bool = kwargs.get("always_include_chapter_numbers", False)

    if (
        _does_reference_include_all_verses_in_start_book(reference)
        and not force_include_chapters
    ):
        return ""

    if is_single_chapter_book(reference.book) and not force_include_chapters:
        return ""

    return f"{reference.start_chapter}:"


def _get_start_verse(reference: NormalizedReference, **kwargs) -> str:
    force_include_chapters: bool = kwargs.get("always_include_chapter_numbers", False)

    if (
        _does_reference_include_all_verses_in_start_book(reference)
        and not force_include_chapters
    ):
        return ""

    return f"{reference.start_verse}"


def _get_end_chapter(
    reference: NormalizedReference, include_chapters: bool = True, **kwargs
) -> str:
    if not include_chapters:
        return ""

    force_include_chapters: bool = kwargs.get("always_include_chapter_numbers", False)

    if reference.end_book and reference.book != reference.end_book:
        if (
            _does_reference_include_all_verses_in_end_book(reference)
            and not force_include_chapters
        ):
            return ""

        if is_single_chapter_book(reference.end_book) and not force_include_chapters:
            return ""

        return f"{reference.end_chapter}:"

    if (
        _does_reference_include_all_verses_in_start_book(reference)
        and not force_include_chapters
    ):
        return ""

    if is_single_chapter_book(reference.book) and not force_include_chapters:
        return ""

    if reference.start_chapter == reference.end_chapter:
        return ""

    return f"{reference.end_chapter}:"


def _get_end_verse(reference: NormalizedReference, **kwargs) -> str:
    force_include_chapters: bool = kwargs.get("always_include_chapter_numbers", False)

    if reference.end_book and reference.book != reference.end_book:
        if (
            _does_reference_include_all_verses_in_end_book(reference)
            and not force_include_chapters
        ):
            return ""

        return f"{reference.end_verse}"

    if (
        _does_reference_include_all_verses_in_start_book(reference)
        and not force_include_chapters
    ):
        return ""

    return (
        f"{reference.end_verse}"
        if reference.start_verse != reference.end_verse
        or reference.start_chapter != reference.end_chapter
        else ""
    )


def _does_reference_include_all_verses_in_start_book(reference: NormalizedReference):
    if reference.start_chapter != 1:
        return False

    if reference.start_verse != 1:
        return False

    if reference.end_book and reference.end_book != reference.book:
        return True

    max_chapters = get_number_of_chapters(reference.book)

    if reference.end_chapter != max_chapters:
        return False

    return reference.end_verse == get_number_of_verses(reference.book, max_chapters)


def _does_reference_include_all_verses_in_end_book(reference: NormalizedReference):
    max_chapters = get_number_of_chapters(reference.end_book)

    if reference.end_chapter != max_chapters:
        return False

    return reference.end_verse == get_number_of_verses(reference.end_book, max_chapters)


# TODO - rewrite this to not need a parser
def format_scripture_text(verse_ids: List[int], **kwargs) -> str:
    """
    Returns the formatted scripture text for the given list of verse IDs.

    :param verse_ids: A list of integer verse ids
    :type verse_ids: List[int]
    :return: The formatted scripture text for the verse ids
    :rtype: str
    """
    one_verse_per_paragraph: bool = kwargs.get("one_verse_per_paragraph", False)
    full_title: bool = kwargs.get("full_title", False)
    format_type: str = kwargs.get("format_type", "html")
    include_verse_numbers: bool = kwargs.get("include_verse_numbers", True)
    parser: BibleParser = kwargs.get("parser", DEFAULT_PARSER)

    if one_verse_per_paragraph or len(verse_ids) == 1:
        return _format_scripture_text_verse_by_verse(
            verse_ids, parser.version, full_title, format_type, include_verse_numbers
        )

    return _format_scripture_text_with_parser(
        verse_ids, parser, full_title, format_type, include_verse_numbers
    )


def _format_scripture_text_verse_by_verse(
    verse_ids: List[int],
    version: Version,
    full_title: bool,
    format_type: str,
    include_verse_numbers: bool,
) -> str:
    verse_ids.sort()
    text: str = ""
    current_book: Optional[Book] = None
    current_chapter: Optional[int] = None

    for verse_id in verse_ids:
        book, chapter_number, verse_number = get_book_chapter_verse(verse_id)

        if book != current_book:
            current_book = book
            current_chapter = chapter_number
            book_titles: Optional[BookTitles] = get_book_titles(book, version)

            if book_titles:
                title: str = (
                    book_titles.long_title if full_title else book_titles.short_title
                )
                text += _format_title(title, format_type, len(text) == 0)

            text += _format_chapter(chapter_number, format_type)

        elif chapter_number != current_chapter:
            current_chapter = chapter_number
            text += _format_chapter(chapter_number, format_type)

        verse_text: Optional[str] = get_verse_text(verse_id, version)

        if include_verse_numbers:
            verse_text = f"{verse_number}. {verse_text}"

        text += _format_paragraph(verse_text, format_type)

    return text


def _format_scripture_text_with_parser(
    verse_ids: List[int],
    parser: BibleParser,
    full_title: bool,
    format_type: str,
    include_verse_numbers: bool,
) -> str:
    title_function: Callable[[Any], Any] = (
        parser.get_book_title if full_title else parser.get_short_book_title
    )
    text: str = ""

    paragraphs: Any = parser.get_scripture_passage_text(
        verse_ids, include_verse_number=include_verse_numbers
    )

    for book, chapters in paragraphs.items():
        title: str = title_function(book)
        text += _format_title(title, format_type, len(text) == 0)

        for chapter, paragraphs in chapters.items():
            text += _format_chapter(chapter, format_type)

            for paragraph in paragraphs:
                text += _format_paragraph(paragraph, format_type)

    return text


def _format_title(title: str, format_type: str, is_first_book: bool) -> str:
    if format_type == "html":
        return f"<h1>{title}</h1>\n"

    if not is_first_book:
        return f"\n\n{title}\n\n"

    return f"{title}\n\n"


def _format_chapter(chapter: int, format_type: str) -> str:
    if format_type == "html":
        return f"<h2>Chapter {chapter}</h2>\n"

    return f"Chapter {chapter}\n\n"


def _format_paragraph(paragraph: Optional[str], format_type: str) -> str:
    if format_type == "html":
        return f"<p>{paragraph}</p>\n"

    return f"   {paragraph}\n"


@lru_cache()
def get_verse_text(verse_id: int, version: Version = DEFAULT_VERSION) -> Optional[str]:
    """
    Returns the scripture text of the given verse id and version of the Bible.

    :param verse_id: a verse id
    :type verse_id: int
    :param version: a version of the Bible, defaults to American Standard
    :type version: :ref:`Version`
    :return: The scripture text of the given verse id and version
    :rtype: str
    :raises InvalidVerseError: if the given verse id does not correspond to a valid verse
    :raises MissingVerseFileError: if the verse file for the given verse_id and version does not exist
    """
    if verse_id not in VERSE_IDS:
        raise InvalidVerseError(verse_id=verse_id)

    try:
        version_verse_texts: Dict[int, str] = _get_version_verse_texts(version)
    except FileNotFoundError as e:
        raise MissingVerseFileError(e)

    return version_verse_texts.get(verse_id)


@lru_cache()
def _get_version_verse_texts(version: Version) -> Dict[int, str]:
    verse_texts: Optional[Dict[int, str]] = VERSE_TEXTS.get(version)

    if verse_texts is None:
        json_filename: str = os.path.join(
            os.path.join(DATA_FOLDER, version.value.lower()), "verses.json"
        )
        verse_texts = {}

        with open(json_filename, "r") as json_file:
            for verse_id, verse_text in json.load(json_file).items():
                verse_texts[int(verse_id)] = verse_text

        VERSE_TEXTS[version] = verse_texts

    return verse_texts


@lru_cache()
def get_book_titles(
    book: Book, version: Version = DEFAULT_VERSION
) -> Optional[BookTitles]:
    """
    Returns the book titles for the given :ref:`Book` and optional :ref:`Version`

    :param book: a book of the Bible
    :type book: :ref:`Book`
    :param version: a version of the Bible, defaults to American Standard
    :type version: :ref:`Version`
    :return: the long and short titles of the given book and version
    :rtype: Optional[BookTitles]
    :raises MissingBookFileError: if the book file for the given book and version does not exist
    """
    try:
        version_book_tiles: Dict[Book, BookTitles] = _get_version_book_titles(version)
    except FileNotFoundError as e:
        raise MissingBookFileError(e)

    return version_book_tiles.get(book)


@lru_cache()
def _get_version_book_titles(version: Version) -> Dict[Book, BookTitles]:
    book_titles: Optional[Dict[Book, BookTitles]] = BOOK_TITLES.get(version)

    if book_titles is None:
        json_filename: str = os.path.join(
            os.path.join(DATA_FOLDER, version.value.lower()), "books.json"
        )
        book_titles = {}

        with open(json_filename, "r") as json_file:
            for book_id, titles in json.load(json_file).items():
                book: Book = Book(int(book_id))
                book_titles[book] = BookTitles(titles[0], titles[1])

        BOOK_TITLES[version] = book_titles

    return book_titles



from typing import Dict, List

import pythonbible as bible


def test_get_references(text_with_reference: str) -> None:
    # Given a text string with at least one scripture reference
    # When parsing that text
    references: List[bible.NormalizedReference] = bible.get_references(
        text_with_reference
    )

    # Then the references are found and returned in a list of normalized reference tuples
    assert len(references) == 2
    assert references[0] == bible.NormalizedReference(
        bible.Book.MATTHEW, 18, 12, 18, 14
    )
    assert references[1] == bible.NormalizedReference(bible.Book.LUKE, 15, 3, 15, 7)


def test_get_references_complex(
    text_with_reference_complex: str,
    normalized_references_complex: List[bible.NormalizedReference],
) -> None:
    # Given a text string with multiple complex references
    # When parsing that text
    references: List[bible.NormalizedReference] = bible.get_references(
        text_with_reference_complex
    )

    # Then the references are found and returned in a list of normalized reference tuples
    assert references == normalized_references_complex


def test_normalize_reference(non_normalized_reference: str) -> None:
    # Given a non-normalized reference (just a string)
    # When we normalize that reference
    normalized_references: List[bible.NormalizedReference] = bible.normalize_reference(
        non_normalized_reference
    )

    # Then the reference is returned as a list of tuples with the book enum,
    # start chapter, start verse, end chapter, and end verse
    assert len(normalized_references) == 1
    assert normalized_references[0] == bible.NormalizedReference(
        bible.Book.MATTHEW, 18, 12, 18, 14
    )


def test_normalize_reference_without_verse_numbers(
    reference_without_verse_numbers: str,
) -> None:
    # Given a non-normalized reference that does not contain verse numbers (just book and chapters)
    # When we normalize that reference
    normalized_references: List[bible.NormalizedReference] = bible.normalize_reference(
        reference_without_verse_numbers
    )

    # Then the resulting normalized references contain the proper verse numbers
    assert len(normalized_references) == 1
    assert normalized_references[0] == bible.NormalizedReference(
        bible.Book.EXODUS, 20, 1, 20, 26
    )


def test_normalize_reference_range_without_verse_numbers(
    reference_range_without_verse_numbers: str,
) -> None:
    # Given a non-normalized reference that does not contain verse numbers (just book and chapters)
    # When we normalize that reference
    normalized_references: List[bible.NormalizedReference] = bible.normalize_reference(
        reference_range_without_verse_numbers
    )

    # Then the resulting normalized references contain the proper verse numbers
    assert len(normalized_references) == 1
    assert normalized_references[0] == bible.NormalizedReference(
        bible.Book.GENESIS, 1, 1, 4, 26
    )


def test_get_references_roman_numerals(
    roman_numeral_references: str,
    normalized_references_complex: List[bible.NormalizedReference],
) -> None:
    # Given a text string with multiple references with roman numerals
    # When parsing that text
    references: List[bible.NormalizedReference] = bible.get_references(
        roman_numeral_references
    )

    # Then the references are found and returned in a list of normalized reference tuples
    assert references == normalized_references_complex


def test_philemon_vs_philippians() -> None:
    """Test for https://github.com/avendesora/pythonbible/issues/2!"""
    # Given a text string with a reference in the book of Philemon
    text: str = "Philemon 1:9"

    # When we parse the references from that text
    references: List[bible.NormalizedReference] = bible.get_references(text)

    # Then the parser does not raise an error and returns the appropriate reference
    assert references == [bible.NormalizedReference(bible.Book.PHILEMON, 1, 9, 1, 9)]


def test_book_alternative_names(book_alternative_names) -> None:
    # Given the books of the Bible with their alternative names/abbreviations

    for book, alternative_names in book_alternative_names.items():
        references = bible.get_references(f"{book.title} 1:1-2")

        for alternative_name in alternative_names:
            # When we parse the references with the alternative name
            alternative_references = bible.get_references(f"{alternative_name} 1:1-2")

            # Then the alternative references match the baseline references
            print(alternative_name)
            assert alternative_references == references


def test_cross_book_reference_just_books() -> None:
    # Given a text string with a reference that ranges over multiple books of the Bible
    text: str = "Genesis - Deuteronomy"

    # When we parse the references from that text
    references: List[bible.NormalizedReference] = bible.get_references(text)

    # Then the parser does not raise an error and returns the appropriate reference
    deuteronomy: bible.Book = bible.Book.DEUTERONOMY
    max_chapter: int = bible.get_number_of_chapters(deuteronomy)
    max_verse: int = bible.get_number_of_verses(deuteronomy, max_chapter)
    assert references == [
        bible.NormalizedReference(
            bible.Book.GENESIS, 1, 1, max_chapter, max_verse, deuteronomy
        )
    ]


def test_cross_book_reference_complex() -> None:
    # Given a text string with a complex reference that ranges over multiple books of the Bible
    text: str = "Genesis 1:1-5, 50:3 - Exodus 1:14, 2:3-20:5"

    # When we parse the references from that text
    references: List[bible.NormalizedReference] = bible.get_references(text)

    # Then the parser does not raise an error and returns the appropriate reference
    assert references == [
        bible.NormalizedReference(bible.Book.GENESIS, 1, 1, 1, 5, None),
        bible.NormalizedReference(bible.Book.GENESIS, 50, 3, 1, 14, bible.Book.EXODUS),
        bible.NormalizedReference(bible.Book.EXODUS, 2, 3, 20, 5, None),
    ]


def test_book_group_reference() -> None:
    # Given a text string with a book group reference
    text: str = "This class is a survey of the Old Testament of the Bible."

    # When we parse the references from that text using the optional book_groups parameter
    references: List[bible.NormalizedReference] = bible.get_references(
        text, book_groups=bible.BOOK_GROUPS
    )

    # Then the parser returns the appropriate book group reference
    malachi: bible.Book = bible.Book.MALACHI
    max_chapter: int = bible.get_number_of_chapters(malachi)
    max_verse: int = bible.get_number_of_verses(malachi, max_chapter)
    assert references == [
        bible.NormalizedReference(
            bible.Book.GENESIS, 1, 1, max_chapter, max_verse, malachi
        ),
    ]


def test_book_group_reference_custom() -> None:
    # Given a custom dictionary of book group regular expressions
    book_groups: Dict[str, List[bible.Book]] = {
        "my custom book group": [
            bible.Book.GENESIS,
            bible.Book.EXODUS,
            bible.Book.MATTHEW,
            bible.Book.MARK,
            bible.Book.JUDE,
        ]
    }

    # and a text string containing a custom book group reference
    text: str = "Can you find my custom book group?"

    # When we parse the references from that text
    references: List[bible.NormalizedReference] = bible.get_references(
        text, book_groups=book_groups
    )

    # Then the parser returns the appropriate book group references
    assert references == [
        bible.NormalizedReference(bible.Book.GENESIS, 1, 1, 40, 38, bible.Book.EXODUS),
        bible.NormalizedReference(bible.Book.MATTHEW, 1, 1, 16, 20, bible.Book.MARK),
        bible.NormalizedReference(bible.Book.JUDE, 1, 1, 1, 25, bible.Book.JUDE),
    ]


def test_single_chapter_book_without_chapter_number() -> None:
    # Given a reference string that does not include the chapter number for a book that only has one chapter
    text: str = "Obadiah 3-6"

    # When we parse the references from that text
    references: List[bible.NormalizedReference] = bible.get_references(text)

    # Then the parser returns the appropriate normalized reference with chapter and verse numbers
    assert references == [
        bible.NormalizedReference(bible.Book.OBADIAH, 1, 3, 1, 6, None)
    ]


def test_book_alternative_names_verbum(book_alternative_names_verbum) -> None:
    # Given the books of the Bible with their alternative names/abbreviations

    for book, alternative_names in book_alternative_names_verbum.items():
        references = bible.get_references(f"{book.title} 1:1-2")

        for alternative_name in alternative_names:
            if len(alternative_name.replace(" ", "")) < 3:
                continue

            # When we parse the references with the alternative name
            alternative_references = bible.get_references(f"{alternative_name} 1:1-2")

            # Then the alternative references match the baseline references
            print(alternative_name)
            assert alternative_references == references



from enum import Enum, auto
from typing import Dict, List

from .books import Book


class BookGroup(Enum):
    """BookGroup is an ``Enum`` containing the default Bible book groupings

    :param name: the unique text identifier of the book group
    :type name: str
    :param value: the unique numerical identifier of the book group
    :type value: int
    :param regular_expression: the regular expression used to find mentions of the book group when searching for references
    :type regular_expression: str
    :param books: the list of books included in the book group
    :type books: List[str]
    """

    OLD_TESTAMENT = auto()
    OLD_TESTAMENT_LAW = auto()
    OLD_TESTAMENT_HISTORY = auto()
    OLD_TESTAMENT_POETRY_WISDOM = auto()
    OLD_TESTAMENT_PROPHECY = auto()
    OLD_TESTAMENT_MAJOR_PROPHETS = auto()
    OLD_TESTAMENT_MINOR_PROPHETS = auto()
    NEW_TESTAMENT = auto()
    NEW_TESTAMENT_GOSPELS = auto()
    NEW_TESTAMENT_HISTORY = auto()
    NEW_TESTAMENT_EPISTLES = auto()
    NEW_TESTAMENT_PAUL_EPISTLES = auto()
    NEW_TESTAMENT_GENERAL_EPISTLES = auto()
    NEW_TESTAMENT_APOCALYPTIC = auto()

    @property
    def books(self) -> List[Book]:
        return _BOOK_GROUP_BOOKS.get(self, [])

    @property
    def regular_expression(self) -> str:
        return _BOOK_GROUP_REGULAR_EXPRESSIONS.get(self, "")


_OLD_TESTAMENT_LAW_BOOKS: List[Book] = [
    Book.GENESIS,
    Book.EXODUS,
    Book.LEVITICUS,
    Book.NUMBERS,
    Book.DEUTERONOMY,
]

_OLD_TESTAMENT_HISTORY_BOOKS: List[Book] = [
    Book.JOSHUA,
    Book.JUDGES,
    Book.RUTH,
    Book.SAMUEL_1,
    Book.SAMUEL_2,
    Book.KINGS_1,
    Book.KINGS_2,
    Book.CHRONICLES_1,
    Book.CHRONICLES_2,
    Book.EZRA,
    Book.NEHEMIAH,
    Book.ESTHER,
]

_OLD_TESTAMENT_POETRY_WISDOM_BOOKS: List[Book] = [
    Book.JOB,
    Book.PSALMS,
    Book.PROVERBS,
    Book.ECCLESIASTES,
    Book.SONG_OF_SONGS,
]

_OLD_TESTAMENT_MAJOR_PROPHETS_BOOKS: List[Book] = [
    Book.ISAIAH,
    Book.JEREMIAH,
    Book.LAMENTATIONS,
    Book.EZEKIEL,
    Book.DANIEL,
]

_OLD_TESTAMENT_MINOR_PROPHETS_BOOKS: List[Book] = [
    Book.HOSEA,
    Book.JOEL,
    Book.AMOS,
    Book.OBADIAH,
    Book.JONAH,
    Book.MICAH,
    Book.NAHUM,
    Book.HABAKKUK,
    Book.ZEPHANIAH,
    Book.HAGGAI,
    Book.ZECHARIAH,
    Book.MALACHI,
]

_OLD_TESTAMENT_PROPHECY_BOOKS: List[Book] = (
    _OLD_TESTAMENT_MAJOR_PROPHETS_BOOKS + _OLD_TESTAMENT_MINOR_PROPHETS_BOOKS
)

_OLD_TESTAMENT_BOOKS: List[Book] = (
    _OLD_TESTAMENT_LAW_BOOKS
    + _OLD_TESTAMENT_HISTORY_BOOKS
    + _OLD_TESTAMENT_POETRY_WISDOM_BOOKS
    + _OLD_TESTAMENT_PROPHECY_BOOKS
)

_NEW_TESTAMENT_GOSPELS_BOOKS: List[Book] = [
    Book.MATTHEW,
    Book.MARK,
    Book.LUKE,
    Book.JOHN,
]

_NEW_TESTAMENT_HISTORY_BOOKS: List[Book] = [
    Book.ACTS,
]

_NEW_TESTAMENT_PAUL_EPISTLES_BOOKS: List[Book] = [
    Book.ROMANS,
    Book.CORINTHIANS_1,
    Book.CORINTHIANS_2,
    Book.GALATIANS,
    Book.EPHESIANS,
    Book.PHILIPPIANS,
    Book.COLOSSIANS,
    Book.THESSALONIANS_1,
    Book.THESSALONIANS_2,
    Book.TIMOTHY_1,
    Book.TIMOTHY_2,
    Book.TITUS,
    Book.PHILEMON,
]

_NEW_TESTAMENT_GENERAL_EPISTLES_BOOKS: List[Book] = [
    Book.HEBREWS,
    Book.JAMES,
    Book.PETER_1,
    Book.PETER_2,
    Book.JOHN_1,
    Book.JOHN_2,
    Book.JOHN_3,
    Book.JUDE,
]

_NEW_TESTAMENT_EPISTLES_BOOKS: List[Book] = (
    _NEW_TESTAMENT_PAUL_EPISTLES_BOOKS + _NEW_TESTAMENT_GENERAL_EPISTLES_BOOKS
)

_NEW_TESTAMENT_APOCALYPTIC_BOOKS: List[Book] = [
    Book.REVELATION,
]

_NEW_TESTAMENT_BOOKS: List[Book] = (
    _NEW_TESTAMENT_GOSPELS_BOOKS
    + _NEW_TESTAMENT_HISTORY_BOOKS
    + _NEW_TESTAMENT_EPISTLES_BOOKS
    + _NEW_TESTAMENT_APOCALYPTIC_BOOKS
)

_BOOK_GROUP_BOOKS: Dict[BookGroup, List[Book]] = {
    BookGroup.OLD_TESTAMENT: _OLD_TESTAMENT_BOOKS,
    BookGroup.OLD_TESTAMENT_LAW: _OLD_TESTAMENT_LAW_BOOKS,
    BookGroup.OLD_TESTAMENT_HISTORY: _OLD_TESTAMENT_HISTORY_BOOKS,
    BookGroup.OLD_TESTAMENT_POETRY_WISDOM: _OLD_TESTAMENT_POETRY_WISDOM_BOOKS,
    BookGroup.OLD_TESTAMENT_PROPHECY: _OLD_TESTAMENT_PROPHECY_BOOKS,
    BookGroup.OLD_TESTAMENT_MAJOR_PROPHETS: _OLD_TESTAMENT_MAJOR_PROPHETS_BOOKS,
    BookGroup.OLD_TESTAMENT_MINOR_PROPHETS: _OLD_TESTAMENT_MINOR_PROPHETS_BOOKS,
    BookGroup.NEW_TESTAMENT: _NEW_TESTAMENT_BOOKS,
    BookGroup.NEW_TESTAMENT_GOSPELS: _NEW_TESTAMENT_GOSPELS_BOOKS,
    BookGroup.NEW_TESTAMENT_HISTORY: _NEW_TESTAMENT_HISTORY_BOOKS,
    BookGroup.NEW_TESTAMENT_EPISTLES: _NEW_TESTAMENT_EPISTLES_BOOKS,
    BookGroup.NEW_TESTAMENT_PAUL_EPISTLES: _NEW_TESTAMENT_PAUL_EPISTLES_BOOKS,
    BookGroup.NEW_TESTAMENT_GENERAL_EPISTLES: _NEW_TESTAMENT_GENERAL_EPISTLES_BOOKS,
    BookGroup.NEW_TESTAMENT_APOCALYPTIC: _NEW_TESTAMENT_APOCALYPTIC_BOOKS,
}

_BOOK_GROUP_REGULAR_EXPRESSIONS: Dict[BookGroup, str] = {
    BookGroup.OLD_TESTAMENT: "Old Testament",
    BookGroup.OLD_TESTAMENT_LAW: "Law",
    BookGroup.OLD_TESTAMENT_HISTORY: "History",
    BookGroup.OLD_TESTAMENT_POETRY_WISDOM: "Poetry|Wisdom",
    BookGroup.OLD_TESTAMENT_MAJOR_PROPHETS: "Major Prophets",
    BookGroup.OLD_TESTAMENT_MINOR_PROPHETS: "Minor Prophets",
    BookGroup.OLD_TESTAMENT_PROPHECY: "Prophecy",
    BookGroup.NEW_TESTAMENT: "New Testament",
    BookGroup.NEW_TESTAMENT_GOSPELS: "Gospels",
    BookGroup.NEW_TESTAMENT_HISTORY: "History",
    BookGroup.NEW_TESTAMENT_PAUL_EPISTLES: "Pauline Epistles|Paul's Epistles|Epistles of Paul",
    BookGroup.NEW_TESTAMENT_GENERAL_EPISTLES: "General Epistles",
    BookGroup.NEW_TESTAMENT_EPISTLES: "Epistles",
    BookGroup.NEW_TESTAMENT_APOCALYPTIC: "Apocalyptic",
}

BOOK_GROUPS: Dict[str, List[Book]] = {
    book_group.regular_expression: book_group.books for book_group in BookGroup
}



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['pythonbible',
 'pythonbible.bible',
 'pythonbible.bible.osis',
 'pythonbible.counters']

package_data = \
{'': ['*'],
 'pythonbible': ['.mypy_cache/*',
                 '.mypy_cache/3.9/*',
                 '.mypy_cache/3.9/_typeshed/*',
                 '.mypy_cache/3.9/collections/*',
                 '.mypy_cache/3.9/ctypes/*',
                 '.mypy_cache/3.9/importlib/*',
                 '.mypy_cache/3.9/json/*',
                 '.mypy_cache/3.9/os/*',
                 '.mypy_cache/3.9/pyexpat/*',
                 '.mypy_cache/3.9/pythonbible/*',
                 '.mypy_cache/3.9/pythonbible/bible/*',
                 '.mypy_cache/3.9/pythonbible/bible/osis/*',
                 '.mypy_cache/3.9/pythonbible/counters/*',
                 '.mypy_cache/3.9/xml/*',
                 '.mypy_cache/3.9/xml/etree/*',
                 '.mypy_cache/3.9/xml/parsers/*',
                 '.mypy_cache/3.9/xml/parsers/expat/*',
                 'data/*'],
 'pythonbible.bible': ['data/*', 'data/asv/*', 'data/kjv/*'],
 'pythonbible.bible.osis': ['versions/*']}

extras_require = \
{'all': ['defusedxml >=0.7.1'],
 'dev': ['bandit >=1.7.4',
         'black >=22.1.0',
         'mypy >=0.931',
         'prospector >=1.7.6'],
 'doc': ['furo >=2022.3.4', 'Sphinx >=4.4.0', 'sphinx-toolbox >=2.18.0'],
 'test': ['pytest >=7.0.1', 'pytest-cov >=3.0.0']}

setup(name='pythonbible',
      version='0.7',
      description='pythonbible includes features for parsing texts for scripture references,',
      author='Nathan Patton',
      author_email='npatton@gmail.com',
      url='https://docs.python.bible',
      packages=packages,
      package_data=package_data,
      extras_require=extras_require,
      python_requires='>=3.7',
     )



from typing import List

import pythonbible as bible


def test_count_books_single_book() -> None:
    # Given a list of references
    references: List[bible.NormalizedReference] = bible.get_references("James 1:4-6")

    # When we get the count of books in the references
    number_of_books: int = bible.count_books(references[0])

    # Then the count is correct
    assert number_of_books == 1


def test_count_books_two_books() -> None:
    # Given a list of references
    references: List[bible.NormalizedReference] = bible.get_references(
        "Matthew 19:3 - Mark 6:9"
    )

    # When we get the count of books in the references
    number_of_books: int = bible.count_books(references[0])

    # Then the count is correct
    assert number_of_books == 2


def test_count_books_multiple_books() -> None:
    # Given a list of references
    references: List[bible.NormalizedReference] = bible.get_references(
        "Genesis - Deuteronomy"
    )

    # When we get the count of books in the references
    number_of_books: int = bible.count_books(references[0])

    # Then the count is correct
    assert number_of_books == 5


def test_count_books_multiple_references() -> None:
    # Given a list of references
    references: List[bible.NormalizedReference] = bible.get_references(
        "Genesis - Deuteronomy, Matthew 19:3 - Mark 6:9, James 1:4-6"
    )

    # When we get the count of books in the references
    number_of_books: int = bible.count_books(references)

    # Then the count is correct
    assert number_of_books == 5 + 2 + 1


def test_count_books_string() -> None:
    # Given a string containing one or more Scripture references
    reference: str = "Genesis - Deuteronomy, Matthew 19:3 - Mark 6:9, James 1:4-6"

    # When we get the count of books in the references
    number_of_books: int = bible.count_books(reference)

    # Then the count is correct
    assert number_of_books == 5 + 2 + 1


def test_count_chapters_single_chapter() -> None:
    # Given a list of references
    references: List[bible.NormalizedReference] = bible.get_references("James 1:4-6")

    # When we get the count of chapters in the references
    number_of_chapters: int = bible.count_chapters(references[0])

    # Then the count is correct
    assert number_of_chapters == 1


def test_count_chapters_two_chapters() -> None:
    # Given a list of references
    references: List[bible.NormalizedReference] = bible.get_references("James 1-2")

    # When we get the count of chapters in the references
    number_of_chapters: int = bible.count_chapters(references[0])

    # Then the count is correct
    assert number_of_chapters == 2


def test_count_chapters_multiple_chapters() -> None:
    # Given a list of references
    references: List[bible.NormalizedReference] = bible.get_references("James")

    # When we get the count of chapters in the references
    number_of_chapters: int = bible.count_chapters(references[0])

    # Then the count is correct
    assert number_of_chapters == 5


def test_count_chapters_multiple_books() -> None:
    # Given a list of references
    references: List[bible.NormalizedReference] = bible.get_references(
        "Matthew 28:1 - Luke 1:10"
    )

    # When we get the count of chapters in the references
    number_of_chapters: int = bible.count_chapters(references[0])

    # Then the count is correct
    assert number_of_chapters == 1 + 16 + 1


def test_count_chapters_multiple_references() -> None:
    # Given a list of references
    references: List[bible.NormalizedReference] = bible.get_references(
        "Genesis, Matthew - Acts"
    )

    # When we get the count of chapters in the references
    number_of_chapters: int = bible.count_chapters(references)

    # Then the count is correct
    assert number_of_chapters == 50 + 28 + 16 + 24 + 21 + 28  # 167 total


def test_count_chapters_string() -> None:
    # Given a string containing one or more Scripture references
    reference: str = "Genesis, Matthew - Acts"

    # When we get the count of chapters in the reference
    number_of_chapters: int = bible.count_chapters(reference)

    # Then the count is correct
    assert number_of_chapters == 50 + 28 + 16 + 24 + 21 + 28  # 167 total


def test_count_verses_single_verse() -> None:
    # Given a list of references
    references: List[bible.NormalizedReference] = bible.get_references("Genesis 1:1")

    # When we get the count of verses in the references
    number_of_verses: int = bible.count_verses(references[0])

    # Then the count is correct
    assert number_of_verses == 1


def test_count_verses_multiple_verses() -> None:
    # Given a list of references
    references: List[bible.NormalizedReference] = bible.get_references("Genesis 1:6-10")

    # When we get the count of verses in the references
    number_of_verses: int = bible.count_verses(references[0])

    # Then the count is correct
    assert number_of_verses == 5


def test_count_verses_multiple_chapters() -> None:
    # Given a list of references
    references: List[bible.NormalizedReference] = bible.get_references(
        "Matthew 5:3-7:27"
    )

    # When we get the count of verses in the references
    number_of_verses: int = bible.count_verses(references[0])

    # Then the count is correct
    assert number_of_verses == 46 + 34 + 27


def test_count_verses_multiple_books() -> None:
    # Given a list of references
    references: List[bible.NormalizedReference] = bible.get_references("1 John - Jude")

    # When we get the count of verses in the references
    number_of_verses: int = bible.count_verses(references[0])

    # Then the count is correct
    assert number_of_verses == (10 + 29 + 24 + 21 + 21) + 13 + 14 + 25


def test_count_verses_multiple_references() -> None:
    # Given a list of references
    references: List[bible.NormalizedReference] = bible.get_references(
        "Genesis 1:1; John 3:16; Romans 15:5-7,13"
    )

    # When we get the count of verses in the references
    number_of_verses: int = bible.count_verses(references)

    # Then the count is correct
    assert number_of_verses == 1 + 1 + (3 + 1)


def test_count_verses_string() -> None:
    # Given a string containing one or more Scripture references
    reference: str = "Genesis 1:1; John 3:16; Romans 15:5-7,13"

    # When we get the count of verses in the reference
    number_of_verses: int = bible.count_verses(reference)

    # Then the count is correct
    assert number_of_verses == 1 + 1 + (3 + 1)
